import logging, traceback, json, re, time

def _median(data):
        """
        Returns the median from a set of data.
        Is not O(1) for memory usage but whatever
        """
        if not data:
            return None
        count = {}
        for item in data:
            if item not in count:
                count[item] = 0
            count[item] += 1
        
        max_count = max(count.values())
        for item in count:
            if count[item] == max_count:
                return item

class Handler():
    def direct(self, Node, value, _):
        """
        Direct method. Just return the result
        """
        return value
        
    def json(self, Node, value, _):
        """
        Handles json method of polling
        """
        
        try:
            item = json.loads(value)
        except:
            logging.debug(value)
            raise
        
        if 'key' not in Node.dict:
            raise ValueError('No key in section, %s' % value)
            
        for key in Node.dict['key'].split(','):
            item = item[key]
            
        if 'strip' in Node.dict and type(item) is str:
        
            item = item.replace(Node.dict['strip'][1:-1], '')
        
        return item
    
    def re(self, Node, resp, _):
        """
        Handles re method of polling
        """
        info = Node.dict
        
        if not resp:
            return
        
        if 'key' not in Node.dict:
            raise ValueError('%s: No key in section' % Node.name) 
       
        #Hacks to deal with wierd parser imports
        Node.dict['key'] = Node.dict['key'].replace('\\r', '\r').replace('\\n', '\n').replace('\\\\d', '\d')
       
        result = re.search( Node.dict['key'], str(resp))
        if not result:
            #raise ValueError('%s: No matching re %s, %s' 
            #        % (Node.name, len(resp), Node))
            return
            
        group = int(Node.dict['group']) if 'group' in Node.dict else 1
        result = result.group(group)
        
        if 'strip' in Node.dict and type(result) is str:
            result = result.replace(Node.dict['strip'][1:-1], '')
        return result
        
    def rate(self, Node, info, source):
        if source == 'rate':
            """
            If this is the rate just update it
            """
            def scale_to_mult(scale):
                scale = scale.lower()
                if scale == 'gh':
                    return 1000**3
                elif scale == 'mh':
                    return 1000 ** 2
                elif scale == 'kh':
                    return 1000
                return 1
            
            rate = float(info)
            mode = Node.namespace.get_node('rate').dict['scale']
            
            hashs = rate * scale_to_mult(mode)
            Node.dict['rate'] = hashs
            return Node.dict.get('value', None)
            
        if source[0:4] == 'time':
            """
            If this is the increment timer increase the shares
            """
            
            #Figure out how long it has been
            old_time = float(Node.dict.get('last_called', time.time()))
            new_time = float(time.time())
            diff = new_time - old_time
            Node.dict['last_called'] = new_time
            
            #Add to shares
            shares = float(Node.dict.get('value', 0))
            rate = float(Node.dict.get('rate', 0))
            new = shares + float(rate)/(2**32) * diff
            return new

        
    def rateduration(self, Node, value, resp):
        if resp == 'rate':
            """
            If this is the rate just update it
            """
            self.rate(Node, value, resp)
            
        elif resp == 'duration':
            """
            If this is duration check for a drop
            """
            if 'rate' not in Node.dict:
                return
            
            rate = float(Node.dict['rate'])
            
            old_dur = Node.dict.get('duration', 0)
            Node.dict['duration'] = float(value)
            
            if float(value) < old_dur:
                return 0
            else:
                diff = float(value) - float(old_dur)
                return float(Node.dict.get('value', 0)) + float(rate)/(2**32) * float(diff)
            
    def difficulty(self, Node, _, __):
         #Difficulty Sites
        if 'coin' not in Node.dict:
            raise ValueError("No coin in node")
            
        coin = Node.dict['coin']
        diffs = []
        for site in Node.objects.difficulty_sites:
            if site[coin]:
                try:
                    float(site[coin])
                except:
                    continue
                diffs.append(site[coin])
        return _median(diffs)
        
    def exchange(self, Node, _, __):
        #Exchange Sites
        
        if 'coin' not in Node.dict:
            raise ValueError("No coin in node")
            
        coin = Node.dict['coin']
        exchange = []
        for site in Node.objects.exchanges:
            if getattr(site, coin, None):
                exchange.append(getattr(site, coin, None))
           
        exchange = _median(exchange)
        if not exchange and Node.name == 'btc':
            return '1.0'
        return exchange
        
        
    def duration(self, Node, value, __):
        def prefix_multiplier(prefix):
            if prefix == 'day':
                return 24*60*60
            if prefix == 'hour':
                return 60*60
            if prefix == 'min':
                return 60
            if prefix == 'sec':
                return 1
            return 1
            
        if 'key' not in Node.dict:
            raise ValueError("No key in node")
        if 'items' not in Node.dict:
            raise ValueError("No items in node")
            
        result = re.search(Node.dict['key'], value)
        
        if not result:
            return
            
        times = Node.dict['items'].split(',')
        
        duration = 0
        index = 0
        for prefix in times:
            index += 1
            try:
                duration += float(result.group(index)) * prefix_multiplier(prefix)
            except TypeError:
                logging.debug('Potential type error %s, %s' % (result.group(index), traceback.format_exc()))
        return duration
        
        
handler = Handler()

def handle(Node, value, source):
    """
    Returns a processed value.
    """
    if 'method' not in Node.dict:
        raise ValueError('No method in node')
    method = Node.dict['method']
    
    func = getattr(handler, method, None)
    if not func:
        raise ValueError('Not a valid function %s' % method)
        
    try:
        return func(Node, value, source)
    except:
        logging.debug(traceback.format_exc())
        logging.debug(Node)
        return None



"""
Represents sites that give us exchange information
"""
    
import baseobject

class Exchange(baseobject.Base_Object):
    """
    Class representing exchanges
    """
        
    def __repr__(self):
        return '<Exchange Site %s, %s>' % (self.name, self.dict)



"""
Api Class which handles basic manipulation of objects
"""
import objects
import pools

class API():
    """
    Wrapper Class for api access
    """
    def __init__(self):
        self.objects = objects.Objects()
        
    def add_pools(self, filenames):
        """
        Adds a set of pools
        """
        for file_path in filenames:
            pool = pools.Pool(file_path, self.objects )
            self.objects.pools.add(pool)
        
    def get_coins(self):
        """
        Returns a set of cointypes
        """
        return self.objects.coins
        
    def get_pools(self):
        """
        Returns a set of pools
        """
        return self.objects.pools
        
    def get_difficulty(self, coin_name):
        """
        Returns the difficulty for the coin
        """
        for coin in self.objects.coins:
            if coin.name == coin_name:
                return coin.difficulty
                
        return None
        
    def get_exchange(self, coin_name):
        """
        Returns the difficulty for the coin
        """
        for coin in self.objects.coins:
            if coin.name == coin_name:
                return coin.exchange
                
        return None
        



import gevent, logging, httplib2, traceback, node_handle

class Node_NameSpace():
    """
    Class representing a name space of nodes
    Used primarily for name resultion
    """
    
    def __init__(self):
        self.nodes = {}
        
    def get_node(self, name):
        if name in self.nodes:
            return self.nodes[name]
        return None
        
    def add_node(self, node):
        if node.name in self.nodes:
            raise KeyError('Duplicate Name')
        self.nodes[node.name] = node
        
    def __repr__(self):
        return '<Node_NameSpace %s>' % (self.nodes)
    

class Node():
    """
    A class representing a node in the information processing chain
    Takes in a dictionary where source is the source node
    Has a set of hooks which are functions to be called on function change
    Also has the ability to produce a dictionary describing its internal state
    """
    
    def __init__(self, name, source_dict, namespace, objects):
        self.name = name
        self.dict = source_dict
        self.hooks = set()
        self.namespace = namespace
        self.namespace.add_node(self)
        self._set_source()
        self.objects = objects
        
    def get_dict(self):
        return self.dict
        
    def _set_source(self):
        """
        Add ourselves to the update list
        """
        if 'source' not in self.dict:
            return
            
        source_addr = self.dict['source']
        for item in source_addr.split(','):
            
            source = self.namespace.get_node(item)
            #If this is a web address make a dummy node
            if not source and 'http' in item[0:4]:
                Http_Node(item, self.namespace)
                source = self.namespace.get_node(item)
                
            if not source and 'time:' in item[0:5]:
                Timer_Node(item, self.namespace)
                source = self.namespace.get_node(item)
                
            if source:
                source.add_hook(self._update_hook)
            else:
                gevent.spawn(self._looping_add, item)
                
    def _looping_add(self, source):
        while not self.namespace.get_node(source):
            gevent.sleep(0.1)
        self.namespace.get_node(source).add_hook(self._update_hook)
            
    def add_hook(self, func):
        "Add someone to the hook list"
        self.hooks.add(func)
        if 'value' in self.dict:
            gevent.spawn(func, self.dict['value'], self.name)
            
    def _trigger(self):
        "Function to trigger everything in hooks"
            
        for func in self.hooks:
            gevent.spawn(func, self.dict['value'], self.name)
        
    def _update_hook(self, value, name):
        """
        Called when our source value changes
        """
        output = node_handle.handle(self, value, name)
        if output == None:
            return
        old = getattr(self.dict, 'value', None)
        self.dict['value'] = output
        if old != output:
            self._trigger()
            
    def __repr__(self):
        return '<Node %s, %s, %s>' % (self.name, self.dict, self.hooks)
        
        
class Http_Node(Node):
    def __init__(self, addr, namespace):
        self.name = addr
        self.hooks = set()
        self.namespace = namespace
        self.namespace.add_node(self)
        self.dict = {}
        gevent.spawn(self._poll)
        
    def _poll(self):
        """
        Looping functions that pulls values
        """
        Http = httplib2.Http(disable_ssl_certificate_validation=True)
        while True:
            try:
                headers, content = Http.request(self.name, headers={})#'Connection':'keep-alive'})
                if content == None:
                    continue
                if len(content) == 0:
                    continue
                if content != self.dict.get('value', None):
                    self.dict['value'] = content
                    self._trigger()
            except:
                logging.debug(traceback.format_exc())
                logging.debug(self.name)
            
            gevent.sleep(60)
            
class Timer_Node(Node):
    def __init__(self, addr, namespace):
        self.name = addr
        self.time = int(addr.split(':')[1])
        self.hooks = set()
        self.namespace = namespace
        self.namespace.add_node(self)
        self.dict = {'value':self.name}
        gevent.spawn(self._poll)
        
    def _poll(self):
        """
        Looping functions that pulls values
        """
        while True:
            self._trigger()
            gevent.sleep(self.time)
                    
import unittest
class TestNodes(unittest.TestCase):
    def setUp(self):
        import gevent.monkey
        gevent.monkey.patch_all(thread = False, time=False)

    def testNamespace(self):
        namespace = Node_NameSpace()
        node1 = Node('test_node', {'source':'http://github.com'}, namespace)
        self.assertIsNot(namespace.get_node('test_node'), None)
        self.assertIsNot(namespace.get_node('http://github.com'), None)
        addr = namespace.get_node('http://github.com')
        addr.dict['value'] = 'test'
        addr._trigger()
        gevent.sleep(0)
        
        value = node1.dict.get('value', None)
        self.assertEquals(value, 'test')
           
if __name__ == '__main__':
    unittest.main()
            



"""
File for describing sites we scrape difficulty off of.
"""

import baseobject

class Site(baseobject.Base_Object):
    """
    Class for difficulty sites.
    """
            
    def __repr__(self):
        return '<Difficulty Site %s>' % (self.name)
        
        
import unittest
class TestNodes(unittest.TestCase):
    def setUp(self):
        import gevent.monkey
        gevent.monkey.patch_all(thread = False, time=False)

    def testNamespace(self):
        pass
           
if __name__ == '__main__':
    unittest.main()



"""
Module to contain pools which have share counts and can be mined
"""
import baseobject

class Pool(baseobject.Base_Object):
    """
    Class of pool that can be mined.
    """
        
    def _setup_general(self, section):
        """
        Sets up general information
        """
        self.general_info = section
        self.name = section.get('name', 'Unknown')
        self.coin = section.get('coin', 'btc')
        self.url = section.get('url', '')
        self.fields.add('name')
        self.fields.add('coin')
        self.fields.add('url')
        
    def __repr__(self):
        return '<Pool %s, %s, %s>' % (self.name, self.coin, self.url)
        
    def _poll(self):
        """
        Updates a couple of statistics. Has special handling for duration
        """
        
        for item in ['api', 'ghash', 'duration','shares']: 
            if item not in self.config.sections():
                continue
            value = self.wrapper.handle(dict(self.config.items(item)))
            if value:
                setattr(self, item, value)
                self.fields.add(item)



"""
Script used to import coins from bitHopper
"""

import ConfigParser

"""
parse = ConfigParser.SafeConfigParser()
parse.read('../bitHopper/whatevercoin.cfg')
for item in parse.sections():
    coin_info = dict(parse.items(item))
    for k in coin_info.keys():
        if 'exchange' in k:
            del coin_info[k]
    
    config = ConfigParser.SafeConfigParser()
    config.add_section('general')
    for k,v in coin_info.items():
        config.set('general', k, v)
    with open('./coins/%s' % coin_info['short_name'], 'wb') as configfile:
        config.write(configfile)
"""
"""      
parse = ConfigParser.RawConfigParser()  
parse.read('../bitHopper/diffwebs.cfg')
for item in parse.sections():
    coin_info = dict(parse.items(item))
    if coin_info['get_method'] == 'regexp':
        coin_info['get_method'] = 're'
    config = ConfigParser.RawConfigParser()
    config.add_section('general')
    config.set('general', 'coin', coin_info['coin'])
    config.set('general', 'name', item)
    coin = coin_info['coin']
    config.add_section(coin)
    config.set(coin, 'address', coin_info['url'])
    config.set(coin, 'method', coin_info['get_method'])
    if 'pattern' in coin_info:
        config.set(coin, 'key', coin_info['pattern'])
    with open('./difficulty_sites/%s' % item, 'wb') as configfile:
        config.write(configfile)
        
"""
parse = ConfigParser.RawConfigParser()
parse.read('../bitHopper/pools.cfg')
parse.read('../bitHopper/user.cfg.default')

for item in parse.sections():
    pool_info = dict(parse.items(item))
    if 'api_method' not in pool_info:
        continue
    config = ConfigParser.RawConfigParser()
    config.add_section('general')
    config.set('general', 'coin', pool_info['coin'])
    config.set('general', 'name', item)
    config.set('general', 'payout_scheme', pool_info.get('payout_scheme','unknown'))
    config.set('general', 'payout_mm', str('payout_merged_mining' in pool_info))
    config.add_section('mine')
    config.set('mine','address', pool_info['mine_address'])
    if 'lp_address' in pool_info:
        config.add_section('lp')
        config.set('lp', 'address', pool_info['lp_address'])
        
    config.add_section('shares')
    
    if 're' in pool_info['api_method']:
        method = 're'
    else:
        method = 'json'
    
    for k,v in pool_info.items():
        if 'api' in k and 'duration' not in k and 'hashrate' not in k:
            k = k.replace('api_','')
            if k == 'address':
                k = 'source'
            config.set('shares', k, v)
            
        elif k == 'api_key_mhashrate':
            config.add_section('rate')
            config.set('rate', 'method', method)
            config.set('rate', 'source', pool_info['api_address'])
            config.set('rate', 'key', v)
            config.set('rate', 'scale', 'mh')
        elif k == 'api_key_ghashrate':
            config.add_section('rate')
            config.set('rate', 'method', method)
            config.set('rate', 'source', pool_info['api_address'])
            config.set('rate', 'key', v)
            config.set('rate', 'scale', 'gh')
        elif k == 'api_key_hashrate':
            config.add_section('rate')
            config.set('rate', 'method', method)
            config.set('rate', 'source', pool_info['api_address'])
            config.set('rate', 'key', v)
            config.set('rate', 'scale', 'h')
        elif k == 'api_key_khashrate':
            config.add_section('rate')
            config.set('rate', 'method', method)
            config.set('rate', 'source', pool_info['api_address'])
            config.set('rate', 'key', v)
            config.set('rate', 'scale', 'kh')
            
        elif 'api_key_duration_' in k:
            fields = k.split('_')[3:]
            config.add_section('duration')
            config.set('duration', 'method', 'duration')
            config.set('duration', 'source', pool_info['api_address'])
            config.set('duration', 'items' , ",".join(fields))
            config.set('duration', 'key', v)
            if 'api_key_duration' in pool_info:
                config.set('duration', 'source', 'duration_pre')
                config.add_section('duration_pre')
                config.set('duration_pre', 'source', pool_info['api_address'])
                config.set('duration_pre', 'method', method)
                config.set('duration_pre', 'key', pool_info['api_key_duration'])
                
            
        elif 'duration' in k:
            print k
        elif 'api' in k and 'hashrate' in k:
            print k
            
    shares_info = dict(config.items('shares'))
    for k,v in shares_info.items():
        if k == 'method' and v == 're_rate':
            config.add_section('rate')
            for k,v in shares_info.items():
                config.set('rate', k, v)
                config.remove_option('shares', k)
            
            #This should be checked
            config.set('rate', 'scale', 'gh')
            
            config.set('rate', 'method', 're')
            
            config.set('shares', 'method', 'rate')
            config.set('shares', 'source', 'rate,time:30')
            
        if k == 'method' and v == 're_rateduration':
            config.set('shares', 'source', 'rate,duration')
            config.set('shares', 'method', 'rateduration')
            
        
    with open('./pools/%s' % item, 'wb') as configfile:
        config.write(configfile)



"""
btcnet_info

A package that gives you info on oodles of bitcoin pools
"""


local_api = None
def __patch():
    """
    Extremely magical gevent deffering patch
    """
    global local_api
    if not local_api:
        import gevent.monkey
        gevent.monkey.patch_all(thread = False, time=False)
        import api
        local_api = api.API()
        
def add_pools(filenames):
    """
    Adds pools defined by their filenames
    """
    __patch()
    local_api.add_pools(filenames)

def get_coins():
    """
    Returns a set of coin objects.
    x.fields describes existing fields
    """
    __patch()
    return local_api.get_coins()
    
def get_pools():
    """
    Returns a set of Pool objects.
    x.fields describes existing fields
    """
    __patch()
    return local_api.get_pools()
    
def get_pool(name):
    """
    Returns a matching pool with the name
    Otherwise returns None
    """
    __patch()
    item = filter(lambda x: x.name == name, local_api.get_pools())
    if item == []:
        return None
    return item[0]
    
def get_difficulty(coin):
    """
    Returns the difficulty of a coin
    """
    __patch()
    return local_api.get_difficulty(coin)
    
def get_exchange(coin):
    """
    Returns the exchange rate for the coin
    """
    __patch()
    return local_api.get_exchange(coin)
    

__all__ = ['get_coins', 'get_pools', 'get_difficulty', 'get_pool', 'get_exchange']




import os, os.path, gevent, pools, exchanges, coins, difficulty_sites, sys
import gevent

class Objects(object):
    def __init__(self):
        self.pools = set()
        self.exchanges = set()
        self.difficulty_sites = set()
        self.coins = set()
        self._setup()
    
    def _setup(self):
        """
        Sets up all of our objects correctly
        """
        #Ugly hack to figure out where we are
        try:
            # determine if application is a script file or frozen exe
            if hasattr(sys, 'frozen'):
                FD_DIR = os.path.dirname(sys.executable)
            else:
                FD_DIR = os.path.dirname(os.path.abspath(__file__))
        except:
            FD_DIR = os.curdir
        
        for file_name in os.listdir(os.path.join(FD_DIR,'difficulty_sites')):
            self.difficulty_sites.add(difficulty_sites.Site(
                reduce(os.path.join, [FD_DIR, 'difficulty_sites', file_name]),
                self,
            ))
        gevent.sleep(0)
            
        #Parse config files and create objects
        for file_name in os.listdir(os.path.join(FD_DIR,'coins')):
            self.coins.add(coins.Coin(
                reduce(os.path.join, [FD_DIR, 'coins', file_name, ]),
                self,
            ))
            
        for file_name in os.listdir(os.path.join(FD_DIR,'exchanges')):
            self.exchanges.add(exchanges.Exchange(
                reduce(os.path.join, [FD_DIR, 'exchanges', file_name]),
                self ,
            ))
        
        for file_name in os.listdir(os.path.join(FD_DIR,'pools')):
            self.pools.add(pools.Pool(
                reduce(os.path.join, [FD_DIR, 'pools', file_name]),
                self
            ))
            
        gevent.sleep(0.1)
        
if __name__ == "__main__":
    Objects()
    gevent.sleep(1)



"""
Aggregation class to store information about coins
"""
import baseobject

class Coin(baseobject.Base_Object):
    """
    Class for coins.
    """
    
    def __repr__(self):
        return '<Coin %s, %s>' % (self.name, self.dict)



#!/usr/bin/env python

from setuptools import setup, find_packages
setup(name='btcnet_info',
        version='0.1.2.15',
        description='BitCoin Network Information Library',
        author='Colin Rice',
        author_email='dah4k0r@gmail.com',
        packages=['btcnet_info'],
        include_package_data = True,
        package_data = { "":["*/*"]},
        url='github.com/c00w/btcnet_info',
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'Topic :: System :: Networking',
            'License :: OSI Approved :: MIT License',
            ],
        license = 'MIT Expat License',
        install_requires = ['gevent']
        )



import unittest
import __init__
import coins
import pools
import gevent

class TestStore(unittest.TestCase):
    var = None
    
    @classmethod
    def setUpClass(self):
        __init__.get_coins()
        gevent.sleep(40)
        
    def testCoins(self):
        for item in __init__.get_coins():
            self.assertTrue(type(item) is coins.Coin )
            
    def testPools(self):
        for item in __init__.get_pools():
            self.assertTrue(type(item) is pools.Pool )
            
    def testMineAddress(self):
        valid = 0
        total = 0
        for item in __init__.get_pools():
            total += 1
            if item['mine.address'] != None:
                valid += 1
        print 'Pools with addresses %s/%s' % (valid, total)
            
    def testPolling(self):
        count = len(__init__.get_pools())
        shares = 0
        for item in __init__.get_pools():
            try:
                a = float(item.shares)
                shares += 1
            except:
                print 'Invalid Pool: %s' %( item.name)
                pass
        print
        print "Valid Pools %s/%s" % (shares, count)
        
        count = len(__init__.get_coins())
        shares = 0
        for item in __init__.get_coins():
            if item.exchange:
                shares += 1
        print "Valid Coins w/ Exchanges %s/%s" % (shares, count)
        
        count = len(__init__.get_coins())
        shares = 0
        for item in __init__.get_coins():
            if item.difficulty:
                shares += 1
        print "Valid Coins w/ Difficulty %s/%s" % (shares, count)
        
    def testApiDifficulty(self):
        for coin in __init__.get_coins():
            float(__init__.get_difficulty(coin.name))
            
    def testApiExchange(self):
        for coin in __init__.get_coins():
            if coin.exchange:
                float(__init__.get_exchange(coin.name))
        

if __name__ == "__main__":
    unittest.main()  



"""
Base object that all sites inherit off of.
"""

try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser
    
import gevent, traceback, httplib2, socket, logging
import node, copy, threading, time

class Base_Object(object):
    """
    Base object for all website classes to inherit off of
    """
    def __init__(self, config_file, objects):
        
        self.config = ConfigParser.RawConfigParser()
        if not self.config.read(config_file):
            raise LookupError("No such config file")
        self.namespace = node.Node_NameSpace()
        self.write_nodes = set()
        self.file_name = config_file
        self.lock = threading.Lock()
        self.lock.acquire()
        
        thread = threading.Thread(target=self.write_thread)
        thread.daemon = True
        thread.start()
        
        for section in self.config.sections():
            if section != 'general':
                item = node.Node(section, dict(self.config.items(section)), self.namespace, objects)
                self.write_nodes.add(item)
            else:
                values = dict(self.config.items(section))
                for item in values:
                    setattr(self, item, values[item])
                    
        self.lock.release()
                    
    def write_poll(self):
        while True:
            while not self.lock.acquire(False):
                gevent.sleep(0)
            
            for node in self.writes:
                section = node.name
                for k,v in node.get_dict().items():
                    self.config.set(section, k,v)
                    
            self.lock.release()
            gevent.sleep(60)
                        
    def write_thread(self):
        mode = 'wb'
        while True:
            try:
                fd = open(self.file_name, mode)
            except IOError, e:
                return
            with self.lock:
                try:
                    fd.seek(0)
                    self.config.write(fd)
                except IOError as e:
                    fd.close()
                    fd = open(self.file_name, mode)
            time.sleep(60)
                    
    def __getattr__(self, name):
        node = self.namespace.get_node(name)
        if node and 'value' in node.dict:
            return node.dict['value']
        elif '.' in name:
            node, index = name.split('.', 1)
            node = self.namespace.get_node(node)
            if node and index in node.dict:
                return node.dict[index]
        return None
        
    def __getitem__(self, name):
        node = self.namespace.get_node(name)
        if node and 'value' in node.dict:
            return node.dict['value']
        elif '.' in name:
            node, index = name.split('.', 1)
            node = self.namespace.get_node(node)
            if node and index in node.dict:
                return node.dict[index]
        elif node:
            return node.dict
        return None
        
    
